/********************************************************
 * CO Buffer Heap - Implementation C++
 *
 * Author:      Hatem Nassrat
 * Institute:   Dalhousie University
 * Class:       Algos for Massive Datasets
 * Instructor:  Dr. Norbert Zeh
 *
 * (Buffer heap refrences listed bellow)
 ********************************************************/

#include <stxxl.h>
#include <string>
#include <iostream>
#include <fstream>
#include <vector> //STL Vector
#include <algorithm>
#include <stack>
#include <cassert>
#include <tclap/CmdLine.h> 

#include <ctime>

using namespace std;

//Priority Queue Operation Codes
#define pqelement 0
#define pqupdate 1
#define pqdelete 2
#define pqdeletemin 3
#define pqsink 4

#ifndef BLPG
#define BLPG 4 //Blocks Per Page
#endif
#ifndef PGCH
#define PGCH 8 //Pages_in_cache
#endif
#ifndef BLSZ
#define BLSZ 256*1024 //=256kb Block Size in bytes
#endif

//Priority Queue Options
unsigned long int Memory = 2*1024*1024;
int nbuffers=2;
string infile_name="pq-ops.bin";
int pqstart_level=0;
int pq_base_size=2;

const unsigned long long int inf=std::numeric_limits<unsigned long long int>::max();


unsigned long long int power(int a, int b)
{
    unsigned long long int c=1;
    for(int i=1; i<=b; i++) c*=a;
    return c;
}

struct level_size {
    long long int el_count;
    long long int op_count;
    long long int max_key;
    long long int del_count;

    level_size(){del_count = el_count = op_count = max_key=0;}
};

struct status {
    unsigned long long int tstamp;
    unsigned short int level;
    char opcode;

    status(){tstamp = level = 0;}
};

struct pq_element {
    unsigned long long int id;
    unsigned long long int key;
    status stat;

    pq_element(){id = key = 0;}

    void print_el(ostream& out){
        if(this->stat.opcode>0)
            out << "OPERATION (";
        else if(this->stat.opcode<0)
            out << "DELETED (";
        else
            out << "ELEMENT (";
            
        out <<  this->id << ", " << this->key << ", [" << 
        this->stat.level<<", "<<(int)this->stat.opcode<<", "
        <<this->stat.tstamp<<"])";    
    }

    bool operator < (const pq_element j) const
    {
        return (key<j.key);
    }
    
    bool operator <= (const pq_element j) const
    {
        return (key<=j.key);
    }
    
    bool operator == (const pq_element j) const
    {
        return (id==j.key);
    }

};

//PqCompare to maintain invariants
struct PqCompare
{
	bool operator()(const pq_element i,const pq_element j) const
    {
        if(i.stat.level > j.stat.level) return true;
        else if(i.stat.level < j.stat.level) return false;
        
        if(i.id < j.id) return true;
        else if(i.id > j.id) return false;

        if(i.stat.tstamp < j.stat.tstamp) return true;
        else if(i.stat.tstamp > j.stat.tstamp) return false;
       
        //if they are equal i wins
        return true;
    }

    // min sentinel = value which is strictly smaller that any input element
    static pq_element min_value()
    {
        pq_element e; 
        e.id = 0; 
        e.stat.level = (std::numeric_limits<unsigned short int>::max)();
        e.stat.opcode = pqupdate;
        e.stat.tstamp = 0; 
        e.key = 0;
        return e;
    }

    // max sentinel = value which is strictly larger that any input element
    static pq_element max_value()
    {
        pq_element e; 
        e.id = inf; 
        e.stat.level = 0;
        e.stat.opcode = 4;
        e.stat.tstamp = inf;
        e.key = inf; 
        return e;
    }

};

bool PqKeyCompare (const pq_element i,const pq_element j)
{
    return (i.key < j.key);
}

struct pq_operation {
    char opcode;
    unsigned long long int element_id;
    unsigned long long int priority;
    
    istream& input_op(istream& input_file)
    {
       return input_file.read(reinterpret_cast < char * > (&(this->opcode)), 1).
       read(reinterpret_cast < char * > (&(this->element_id)), 8).
       read(reinterpret_cast < char * > (&(this->priority)), 8);
    }

};

struct IncLevel
{
    void operator() (pq_element & val)
    { val.stat.level++; }
};

ostream& operator<<(ostream& out, pq_element& p);
void print_element(pq_element& p);
ostream& operator<<(ostream& out, level_size &i);
ostream& print_level_stats(ostream& out, level_size &i);
istream& operator>>(istream& in, pq_operation& p); //binary input

//Typedefs
typedef stxxl::VECTOR_GENERATOR<pq_element,BLPG,PGCH,BLSZ>::result vector_type;
//typedef std::vector<pq_element> vector_type;

/************************************************************************
 * Buffer_Heap 
 * 
 *  public vector pq, vector holding all levels of the bufferheap
 *  each element within the vector is labeled with
 *      ->  The level it belongs to
 *      ->  An opcode if it is an operation 
 *          and opcode=0 if element
 *      ->  TimeStamp (relevant only if operation)
 *      ->  Element ID
 *      ->  Element Key (or priority)
 *  levels increase in size exponentialy, and are stored
 *  starting with level r ending with level 0 in the back
 *  of the vector pq
 *
 *  The following invariants are allways true for all levels:
 *      1.  There are at most 2^i elements at level i
 *      2.  There are at most 2^i updates at level i
 *      3.  For i in range[0,r-1) the keys of all
 *          elements in level i is no larger than
 *          the keys of any element in level i+1
 *      4.  For each element x in level i, all updates
 *          applicable to x reside in levels [0,i]
 *      5.  members of the vector pq are kept sorted according
 *          to the following criteria (in that order):
 *          i.   level (descending)
 *          ii.  element_id (ascending)
 *          iii. opcode (operations before elements)
 *          iv.  timestamp (descending)
 *          v.   (if tie at timestamp (impossible) 
 *               then updates before deletes)
 * Note:    this buffer heap is designed from the cache oblivous buffer 
 *          heap discussed in [*] and [**]
 * [*]  Chowdhury & Ramachandran. CO shortest paths in graphs using 
 *      buffer heap. Proceedings of 16th SPAA. pp 245-54, 2004
 * [**] Tong & Ramachandran. Impl and Exp Evaluation of the CO
 *      Buffer Heap. Thesis, May 2006.
 *************************************************************************/
class buffer_heap 
{
    
    public:
        unsigned short int no_levels;
        unsigned long long int global_timestamp;
      //internal memory??? since max levels = 7*8 = 56 to host 2^(8*8) elements
        vector<level_size> isize; 
      //external memory since it holds N elements and Operations where N >> M  
        vector_type pq;
        
        //constructor
        buffer_heap()
        {
            no_levels=1;
            level_size lev0;
            isize.push_back(lev0);
            global_timestamp = 0;
        }

        ostream& printall(ostream& out)
        {
            vector_type::iterator it;
            out << "PQ contains:\n";
            out << "TYPE\t(ID,\tKEY,\t[Lev, OP-Code, TimeStamp])\n";
            stxxl::for_each(pq.begin(),pq.end(),print_element,nbuffers);
            //for ( it=pq.begin() ; it < pq.end(); it++ )
            //   out<<*it;
            return out;
        }


        /*********************************************
         *Update(pq_el)
         *
         * Input: 
         *  pq_element
         *  w/ 
         *      key == newKey to update
         *      id == element_id
         *      stat.tstamp = operation rank
         *      stat.level = root_level
         *      stat.opcode == Update
         *      
         * Return: 
         *  0 if successful
         *  Error_code otherwise
         *********************************************/
        int Update(pq_element p)
        {
            stack<pq_element> B_prime;
            int level = 0;

            if(p.stat.opcode != pqupdate) //error invalid opcode
            {
                p.stat.opcode = pqupdate;
            }
            p.stat.level = level;
            p.stat.tstamp = global_timestamp++;

            //insert op into Updates Buf of root_level
            if(this->isize[0].del_count>0)
            {
                vector_type::iterator j = this->pq.begin()+level_begin_index(0)+this->isize[0].el_count+this->isize[0].op_count;
                j->id = p.id;
                j->key = p.key;
                j->stat.opcode = p.stat.opcode;
                j->stat.level = p.stat.level;
                j->stat.tstamp = p.stat.tstamp;
                this->isize[0].del_count--;
            }
            else
            {
                this->pq.push_back(p);
                this->isize[0].op_count++;
            }

            FixU(level);
            
            // Redistribute B' to shallowest levels (maintain invariants)

            // Reconstuct

            return 0;
        }

        /**********************************************
         *Delete(pq_el): Lazy PQ Delete
         *
         * Input: 
         *  pq_element
         *  w/ 
         *      id == element_id
         *      stat.tstamp = operation rank
         *      stat.level = root_level
         *      stat.opcode == Update
         *      
         * Return: 
         *  0 if added to bh
         *  Error_code otherwise
         *********************************************/
        int Delete(pq_element p)
        {
            int level = 0;
            if(p.stat.opcode != pqupdate) //error invalid opcode
            {
                p.stat.opcode = pqupdate;
            }
            p.stat.level = level;
            p.stat.tstamp = global_timestamp++;

            //insert op into Updates Buf of root_level
            this->pq.push_back(p);
            this->isize[0].op_count++;
            
            FixU(level);
            
            return 0;
        }
        
        unsigned long long int find_min_key(int level)
        {
            while(level<this->no_levels && this->isize[level].el_count==0)level++;
            if(level == this->no_levels) return inf;
            vector_type::iterator begin = this->pq.begin() +
                level_begin_index(level);
            vector_type::iterator end = begin +
                this->isize[level].el_count; //since no ops
            vector_type::iterator it = begin;
            vector_type::iterator min = it;
            
            while(++it<end)
                if(min->key > it->key)
                    min = it; 
            return min->key;        
        }

        /**********************************************
         *DeleteMin()
         *
         * Input: 
         *  reference to pq_element p
         *
         * Return: 
         *  pq_element in p & 0 if successful
         *  return error_code  otherwise
         *********************************************/
        int DeleteMin(pq_element& p)
        {
            int level = -1;
            bool found = false;
            while(level < this->no_levels-1)
            {
                level++;
                ApplyUpdates(level);
                if(this->isize[level].el_count>0) {found = true; break;}
            }
            
            if(!found) return 1;

            vector_type::iterator begin = this->pq.begin() +
                level_begin_index(level);
            vector_type::iterator end = begin +
                this->isize[level].el_count; //since no ops
            vector_type::iterator it = begin;
            vector_type::iterator min = it;
            
            while(++it<end)
                if(min->key > it->key)
                    min = it; 
            
            p.id = min->id;
            p.key = min->key;
            p.stat.level = min->stat.level;
            p.stat.tstamp = min->stat.tstamp;
            p.stat.opcode = min->stat.opcode; //should be pqelement
            
            min->id = inf;
            min->stat.opcode = -1;
            //have to sort in order to fix the el count
            stxxl::sort(begin,end,PqCompare(),Memory);
            this->isize[level].el_count--;
            this->isize[level].del_count++;
            
            //Two options
            //1. redistribute elements in level to shallowest
            //2. Reconstruct = apply and redistribute level by level

            return 0;
        }


        /*********************************************
         * FixU(i,B_prime)
         *
         *  Input: 
         *      set of minimal elements B_prime
         *      level <- i
         *
         *  Return:
         *      0 upon Success
         *      Error_code otherwise
         *********************************************/
         int FixU(int i)
         {
            while(i<no_levels &&
            this->isize[i].op_count>=power(pq_base_size,i+pqstart_level))
            {
               ApplyUpdates(i);
               //redistribute to upper levels
               i++;
            }
            return 0;
         }
        
        unsigned long long int level_begin_index(int i)
        {
            unsigned long long int lev_begin=0;
            unsigned short int j;
                
            for(j=this->no_levels-1;j>i;j--){
                lev_begin+= this->isize[j].op_count+
                    this->isize[j].el_count + this->isize[j].del_count;
            }

            return lev_begin;
        }

        int ApplyUpdates(int i)
        {
            unsigned long long int j;
            
            if(this->isize[i].op_count == 0) 
                return 0; //level empty nothing to do
			
            vector_type::iterator range_begin = pq.begin() + 
				level_begin_index(i);
			vector_type::iterator range_end = range_begin+
				(this->isize[i].op_count+this->isize[i].el_count);
				
			stxxl::sort(range_begin,
                            range_end+this->isize[i].del_count, 
                            PqCompare(), Memory);
							
			
			if(this->isize[i].el_count == 0 && i< this->no_levels-1)
            {   
                    
                stxxl::for_each_m(range_begin,range_end,IncLevel(),nbuffers);
                
                //Set U_i <- 0
                this->isize[i+1].op_count += this->isize[i].op_count;
                this->isize[i].op_count=0;
                
				//operations are now appended to the end of the next
                //level
            }
            else
            {
                
                unsigned long long int k;
                if(i < this->no_levels-1) k=find_min_key(i+1);
                else k = inf;
                
                //traversing backward (next index before j)
                vector_type::iterator j=range_begin;
                vector_type::iterator next=range_begin;
                //elements that are deleted are just marked
                //with opcode = -1 and id=inf
                vector_type::iterator deleted=range_end;
                
                long long int delta_els = 0;
                long long int delta_ops = 0;
                long long int delta_del = 0;
                long long int delta_push_ops = 0;
                //long long int delta_push_els = 0;
               
                //process to make next point to an element
                //and j the next operation
               while(next<range_end) 
                {  
                    if(j<range_end && j != next)
                    {
                        if(j->stat.opcode == pqdelete)
                        {
                            if(next->id != j->id)
                            {
                                //Element to delete not at my level
                                if(i<this->no_levels-1)
                                {
                                    j->stat.level++;
                                    delta_push_ops++;
                                    delta_ops--;
                                    next = ++j; ++j;
                                }
                                else
                                {
                                    //element to del not in PQ
                                    j->id = inf;
                                    j->stat.opcode = -1;
                                    delta_ops--;
                                    delta_del++; 
                                    next = ++j; ++j;
                                }
                            }
                            else
                            {
                                if(next->stat.opcode == pqelement
                                 || next->stat.opcode == pqsink)
                                {
                                    //found the element to delete
                                    //this->pq.erase(j);
                                    j->id = inf;
                                    j->stat.opcode = -1;
                                    //this->pq.erase(next);
                                    next->id = inf;
                                    next->stat.opcode = -1;
                                    //removed both signal and element
                                    if(next->stat.opcode>0) delta_ops-=2;
                                    else {delta_els--; delta_ops--;}
                                    delta_del+=2;
                                    next = ++j; ++j;
                                }
                                else if(next->stat.opcode == pqupdate)
                                {
                                    //I can remove the update since an
                                    //since there is a younger delete 
                                    //in the pq

                                    //this->pq.erase(next);
                                    next->id = inf;
                                    next->stat.opcode = -1;
                                    
                                    delta_ops--;
                                    delta_del++;
                                    next = j; ++j;
                                }
                                else if(next->stat.opcode == pqdelete)
                                {
                                    //impossible two deletes to same element
                                    //erase one
                                    //this->pq.erase(next);
                                    next->id = inf;
                                    next->stat.opcode = -1;
                                    
                                    delta_ops--;
                                    delta_del++;
                                    next = j; ++j;
                                }
                            }
                        }
                    
                        else if(j->stat.opcode == pqupdate) 
                        {
                            if(next->id != j->id)
                            {
                                //Element to update not at my level
                                if(j->key > k) 
                                {
                                    j->stat.level++;
                                    delta_ops--;
                                    delta_push_ops++;
                                    next = ++j; ++j;
                                }
                                else
                                {
                                    //Update becomes element
                                    j->stat.opcode=0;
                                    delta_ops--;delta_els++;
                                    if(i<this->no_levels-1)
                                    {
                                        //update inserts a delete
                                        vector_type::iterator temp;
                                        bool flag = false;
                                        if(deleted <
                                        range_end+this->isize[i].del_count)
                                        {
                                            temp = deleted++;
                                            flag = true;
                                        }
                                        else if (delta_del>0)
                                        {
                                            temp = range_begin;
                                            while(temp<range_end &&
                                            temp->stat.opcode!=-1) temp++;
                                            if(temp<range_end)
                                                flag=true;
                                        }
                                        
                                        if(flag) //empty cell found
                                        {
                                            temp->stat.level++;
                                            temp->stat.opcode = pqdelete;
                                            temp->stat.tstamp = 
                                                ++global_timestamp;
                                            delta_del--;delta_push_ops++;
                                        }//else SOL
                                    }
                                    
                                    if(this->isize[i].max_key<j->key)
                                        this->isize[i].max_key=j->key;
                                    next = j; ++j;
                                }
                            }
                            else
                            {
                                if(next->stat.opcode == pqelement || 
                                   next->stat.opcode == pqsink)
                                {
                                    if(next->key>j->key) next->key = j->key;
                                    //this->pq.erase(j);
                                    j->id = inf;
                                    j->stat.opcode = -1;
                                    delta_del++;
                                    delta_ops--;
                                    j++;
                                }
                                else if(next->stat.opcode == pqdelete)
                                {
                                    //this->pq.erase(next);
                                    next->id = inf;
                                    next->stat.opcode = -1;
                                    delta_del++;
                                    delta_ops--;
                                    next = j; ++j;

                                }
                                else if(next->stat.opcode == pqupdate)
                                {
                                    //impossible two updates to same element
                                    //erase one keeping smaller key
                                    if(next->key>j->key) next->key = j->key;
                                    //this->pq.erase(j);
                                    j->id = inf;
                                    j->stat.opcode = -1;
                                    delta_del++;
                                    delta_ops--;
                                    j++;
                                }
                            }
                        }
                        
                        else if(j->stat.opcode == pqsink)
                        {
                            if(next->id != j->id)
                            {
                                //Element to update not at my level
                                if(j->key > k)
                                {
                                    j->stat.level++;
                                    delta_push_ops++;
                                    delta_ops--;
                                    next = ++j; ++j;
                                }
                                else
                                {
                                    //sink turns into element
                                    j->stat.opcode = pqelement;
                                    delta_els++; delta_ops--;
                                    next = j; ++j;
                                }
                            }
                            else
                            {
                                if(next->stat.opcode == pqelement)
                                {
                                    if(next->key>j->key) next->key = j->key;
                                    //this->pq.erase(j);
                                    j->id = inf;
                                    j->stat.opcode = -1;
                                    delta_del++;
                                    delta_ops--;
                                    j++;
                                }
                                else if(next->stat.opcode == pqdelete)
                                {
                                    //this->pq.erase(next);
                                    next->id = inf;
                                    next->stat.opcode = -1;
                                    delta_del++;
                                    delta_ops--;
                                    next = j; j++;
                                }
                                else if(next->stat.opcode == pqupdate)
                                {//next could not be an op
                                    //relinquish key to update
                                    if(next->key>j->key) next->key = j->key;
                                    //this->pq.erase(j);
                                    j->id = inf;
                                    j->stat.opcode = -1;
                                    delta_del++;
                                    delta_ops--;
                                    j++;
                                }
                                else if(next->stat.opcode == pqsink)
                                {
                                    //impossible two sinks to same element
                                    //erase one keeping smaller key
                                    if(next->key>j->key)  next->key = j->key;
                                    //this->pq.erase(j);
                                    j->id = inf;
                                    j->stat.opcode = -1;
                                    delta_del++;
                                    delta_ops--;
                                    j++;
                                }
                            }
                        }
                        //else j is an element
                        else next=j++;
                        
                    }else
                    {
                        if(next->stat.opcode==pqdelete)
                        {
                            //last operation on level
                            next->stat.level++;
                            delta_ops--;
                            delta_push_ops++;
                        }
                        else if(next->stat.opcode== pqupdate)
                        {
                            if(next->key > k) 
                            {
                                next->stat.level++;
                                delta_ops--;
                                delta_push_ops++;
                            }
                            else
                            {
                                //Update becomes element
                                next->stat.opcode=0;
                                delta_ops--;delta_els++;
                                if(i<this->no_levels-1)
                                {
                                    //update inserts a delete
                                    vector_type::iterator temp;
                                    bool flag = false;
                                    if(deleted <
                                    range_end+this->isize[i].del_count)
                                    {
                                        temp = deleted++;
                                        flag = true;
                                    }
                                    else if (delta_del>0)
                                    {
                                        temp =   range_begin;
                                        while(temp<range_end &&
                                        temp->stat.opcode!=-1) temp++;
                                        if(temp<range_end) flag=true;
                                    }
                                    
                                    if(flag)
                                    {
                                        temp->stat.level++;
                                        temp->stat.opcode = pqdelete;
                                        temp->stat.tstamp = 
                                            ++global_timestamp;
                                        delta_del--;delta_push_ops++;
                                    }//else SOL
                                }
                                
                                if(this->isize[i].max_key<j->key)
                                    this->isize[i].max_key=j->key;

                            }
                         
                        }
                        else if(next->stat.opcode == pqsink)
                        {
                            if(next->key > k) 
                            {
                                next->stat.level++;
                                delta_ops--;delta_push_ops++;
                            }
                            else
                            {
                                //sink becomes element
                                next->stat.opcode = pqelement;
                                delta_els++; delta_ops--;
                            }
 
                        }

                        if(j==next) j++;
                        else next++; //end of the line
                    }
                    
                }

				//sort entire level
                stxxl::sort(range_begin,
                            range_end+this->isize[i].del_count, 
                            PqCompare(), Memory);
               
                //fix level parameters
                this->isize[i].el_count+=delta_els;
                this->isize[i].op_count+=delta_ops; //should add up to zero
                this->isize[i].del_count+=delta_del;
                if(delta_push_ops>0) 
                    this->isize[i+1].op_count += delta_push_ops;
                
				if(this->isize[i].el_count>power(pq_base_size,i+pqstart_level))
                {
                    //buffer overflow
					
                    //range_begin = this->pq.begin()+level_begin_index(i);
                    range_end = range_begin+this->isize[i].el_count;
					
                    nth_element(range_begin,
                        range_begin+ (power(pq_base_size,i+pqstart_level)-1),
                        range_end);
                    
                    //push rest of elements to lower level as sink
                    delta_els = 0;
                    delta_push_ops = 0;
                    //find the new maximum key
                    j = range_begin;
                    this->isize[i].max_key = 0;
                    while(j<range_begin+power(pq_base_size,i+pqstart_level))
                    {
                        if(j->key > this->isize[i].max_key)
                            this->isize[i].max_key = j->key;
                        j++;
                    }
                   //here j would be at the correct position 
                    while (j<range_end)
                    {
                        j->stat.opcode = pqsink;
                        j->stat.level++;
                        delta_els--;
                        delta_push_ops++;
                        ++j;
                    }
                    this->isize[i].el_count +=delta_els;
                    if(i == this->no_levels-1) 
                    {
                        //add an extra level;
                        this->no_levels++;
                        level_size ls;
                        ls.op_count = delta_push_ops;
                        this->isize.push_back(ls);
                    }
                    else this->isize[i+1].op_count+= delta_push_ops;
                    delta_del = 0;
                    //check for level overflow
                    //occurs when carying deleted elements
                    while(this->isize[i].el_count +
                            this->isize[i].del_count + delta_del >
                            2*power(pq_base_size,i+pqstart_level))
                    {
                        j++->stat.level++;
                        delta_del--;
                    }
                   
                    //ready to regain invariants after nth_elment
                    
                    stxxl::sort(range_begin,range_end+this->isize[i].del_count,
                            PqCompare(), Memory);
                    
                    //fix the counts    
                    this->isize[i].del_count += delta_del;
                    if(i == this->no_levels-1) 
                    {
                        //add an extra level;
                        this->no_levels++;
                        level_size ls;
                        ls.del_count = -delta_del;
                        this->isize.push_back(ls);
                    }
                    else this->isize[i+1].del_count -= delta_del;


                }
                
            }
            
            //cerr<<"STOP APPLY("<<i<<") "<<this->isize[i];

            return 0;
        }

    //end of public members of buffer_heap    
};

ostream& operator<<(ostream& out, buffer_heap &b);


int main (int argc, char* argv[]) 
{

time_t t1,t2; 

    ifstream input_file;
    buffer_heap bh;
    pq_element element;
    pq_operation operation;


try {  

        TCLAP::CmdLine cmd("IO EFFECIENT PQ - Hatem Nassrat", ' ', "");

        TCLAP::ValueArg<std::string> file("i","infile","infile to process", 
                true,"","string");
        cmd.add( file );
        
        TCLAP::ValueArg<int> start("s","startl",
                "Buffer Heap Start Level Default (0)", 
                false, pqstart_level,"int");
        cmd.add( start );

        TCLAP::ValueArg<int> base("b",
                "base","Base Size of Buffers Default (2)"
                , false, pq_base_size,"int");
        cmd.add( base );
        
        TCLAP::ValueArg<int> mem_cap("m",
                "memcap","Aditional Memory Cap for Sorting in MB (2)"
                , false, 2 ,"int");
        cmd.add( mem_cap );

        cmd.parse( argc, argv );

        infile_name = file.getValue();
        pqstart_level = start.getValue();
        pq_base_size = base.getValue();
        Memory = mem_cap.getValue() * 1024 * 1024;

    } catch (TCLAP::ArgException &e)  // catch any exceptions
    { 
        std::cerr << "error: " << e.error() << " for arg " 
                << e.argId() << std::endl; 
    }
   
    int approx = (BLPG*PGCH*BLSZ + Memory)/(1024*1024); 
    cout<<"Approx Max Memory Usage (MB) = "<< approx <<endl;
    
    unsigned long long int count=0;
    (void) time(&t1);
    input_file.open (infile_name.c_str() , ios::in | ios::binary);
    //input_file.read(reinterpret_cast < char * > (&(operation.opcode)), 1).
    //    read(reinterpret_cast < char * > (&(operation.element_id)), 8).
    //    read(reinterpret_cast < char * > (&(operation.priority)), 8)
    while(input_file >> operation)
    {   count++;
        switch(operation.opcode)
        {
            case pqupdate:
                element.id = operation.element_id;
                element.key = operation.priority;
                element.stat.opcode = operation.opcode;
#if DEBUG
                cerr<<"PQ Update => "<<element<<endl;
#endif
                bh.Update(element);
#if DEBUG
                cerr<<bh<<endl;
#endif
                break;
            case pqdelete:
                element.id = operation.element_id;
                element.stat.opcode = operation.opcode;
#if DEBUG
               cerr<<"PQ Update => "<<element<<endl;
#endif
                bh.Delete(element);
#if DEBUG
                cerr<<bh<<endl;
#endif
                break;
            case pqdeletemin:
#if DEBUG
                cerr<<"PQ DeleteMin CALLED,";
#endif
                //HERE APPLY-UPDATES is called untill a min
                //element is retrieved and returned
#if DEBUG 
                if(bh.DeleteMin(element))
                    cout<<" QUEUE EMPTY\n";
                else
                    cout<<" RETURN = "<<element<<endl;
#else
                bh.DeleteMin(element);    
#endif                    
#if DEBUG
                cerr<<bh<<endl;
#endif
                break;
        }
        //WHEN LEVEL UPDATE BUFFER OVERFLOWS, APPLY-UPDATES(level)
        //LEVEL is allways pqstart_level, 
        //Appplying Updates may cause more overflows

    }
   
    input_file.close();
    (void) time(&t2);    
    cout << "Processed"<<count<<" operations\n";
    cout << "Time Elapsed = "<<(int)(t2-t1)<<" seconds\n";

    return 0;
}

ostream& operator<<(ostream& out, pq_element& p) {
    p.print_el(out);
    return out;
}

void print_element(pq_element& p)
{
    p.print_el(cout); cout<<endl;
}

ostream& operator<<(ostream& out, level_size &i)
{
    return print_level_stats(out,i);
}

ostream& print_level_stats(ostream& out, level_size &i)
{
    out<<"Element Count = "<<i.el_count<<" OP Count = "<<i.op_count
        <<" Deleted Count = "<<i.del_count<<" MaxKey = "<<i.max_key
        <<endl;
    return out;
}

istream& operator>>(istream& in, pq_operation& p) {
    return p.input_op(in);
}

ostream& operator<<(ostream& out, buffer_heap &b)
{
    return b.printall(out);
}
